#!/usr/bin/python3 -u
#
#  AttackTree.py
#
#  ADT for Attack Trees (aka ATs) whose BASes can have attributes for metric computation
#
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
#
#  Distribution license: GPLv3
#                        https://www.gnu.org/licenses/gpl-3.0.en.html
#  Copyright 2024–
#  Authors:
#  - Carlos E. Budde <carlosesteban.budde@unitn.it> (Università di Trento, IT)
#
# ------------------------------------------------------------------------------

import copy
import math
import networkx as nx
import pandas as pd
import matplotlib.pyplot as plt
from os.path import isfile as file_exists
from AttackTree_algorithms import AT_metric, METRICS


def deprecated(fun):
    """ Decorator: deprecated function """

    def inner(*arg):
        print("WARNING: function {} is deprecated and will be removed in the future"
              .format(fun.__name__))
        return fun(*arg)

    return inner


class AttackTree:
    """
    An Attack Tree (aka AT) whose BASes can have attributes for metric computation

    - **AT structure:** Can define both tree- and DAG-structured ATs
    - **Gates types:** ATs can be static (OR/AND gates) or dynamic (also SAND gates)
    - **Metrics:** Offers the following metric computations (*for static ATs only*):
        - Probability of attack
        - Cost of ...
    - **MITRE ATT&CK**: BASes can be added to match a *technique* and *tactic*
      from MITRE, which assigns a probability automatically
    """

    # Tactics are id'd by order in these lists: KEEP THE ORDER!!!
    _internal_tactic_ID = [
        "reco",
        "reso",
        "init",
        "exec",
        "pers",
        "priv",
        "defe",
        "cred",
        "disc",
        "late",
        "coll",
        "comm",
        "exfi",
        "impa",
    ]
    _MITRE_tactic_ID = [
        "TA0043",  # "reco"
        "TA0042",  # "reso"
        "TA0001",  # "init"
        "TA0002",  # "exec"
        "TA0003",  # "pers"
        "TA0004",  # "priv"
        "TA0005",  # "defe"
        "TA0006",  # "cred"
        "TA0007",  # "disc"
        "TA0008",  # "late"
        "TA0009",  # "coll"
        "TA0011",  # "comm"
        "TA0010",  # "exfi"
        "TA0040",  # "impa"
    ]
    _MITRE_tactic_names = [
        "Reconnaissance",  # "reco"
        "Resource Development",  # "reso"
        "Initial Access",  # "init"
        "Execution",  # "exec"
        "Persistence",  # "pers"
        "Privilege Escalation",  # "priv"
        "Defense Evasion",  # "defe"
        "Credential Access",  # "cred"
        "Discovery",  # "disc"
        "Lateral Movement",  # "late"
        "Collection",  # "coll"
        "Command and Control",  # "comm"
        "Exfiltration",  # "exfi"
        "Impact",  # "impa"
    ]
    _MITRE_tactic_codenames = [
        "reconnaissance",  # "reco"
        "resource-development",  # "reso"
        "initial-access",  # "init"
        "execution",  # "exec"
        "persistence",  # "pers"
        "privilege-escalation",  # "priv"
        "defense-evasion",  # "defe"
        "credential-access",  # "cred"
        "discovery",  # "disc"
        "lateral-movement",  # "late"
        "collection",  # "coll"
        "command-and-control",  # "comm"
        "exfiltration",  # "exfi"
        "impact",  # "impa"
    ]
    _MITRE_tactics = {
        "reco": ["TA0043", "Reconnaissance", "reconnaissance", ],
        "reso": ["TA0042", "Resource Development", "resource-development", ],
        "init": ["TA0001", "Initial Access", "initial-access", ],
        "exec": ["TA0002", "Execution", "execution", ],
        "pers": ["TA0003", "Persistence", "persistence", ],
        "priv": ["TA0004", "Privilege Escalation", "privilege-escalation", ],
        "defe": ["TA0005", "Defense Evasion", "defense-evasion", ],
        "cred": ["TA0006", "Credential Access", "credential-access", ],
        "disc": ["TA0007", "Discovery", "discovery", ],
        "late": ["TA0008", "Lateral Movement", "lateral-movement", ],
        "coll": ["TA0009", "Collection", "collection", ],
        "comm": ["TA0011", "Command and Control", "command-and-control", ],
        "exfi": ["TA0010", "Exfiltration", "exfiltration", ],
        "impa": ["TA0040", "Impact", "impact", ],
    }

    def __init__(self):
        self._AT = nx.DiGraph()
        self._idx = 0
        self._root = None
        self._BASes = {}
        self._gates = {}
        self._techtac = {}
        self._num_nodes = 0
        self._gates_types = ["OR", "AND", "SAND"]
        self._gates_shapes = dict(zip(self._gates_types, ["v", "^", ">"]))  # for plotting
        self._gates_shapes2 = dict(zip(self._gates_types, ["invtriangle", "triangle", "rarrow"]))  # for plotting
        self._node_shapes = []  # for plotting
        self._node_labels = {}  # "
        self._node_colors = []  # "
        self._MITRE_techniques = None
        self._MITRE_techniques_per_tactic = None
        # Try to load data about MITRE techniques probabilities
        MITRE_techniques_overall = "./data/techniques_frequency_overall.csv"
        if file_exists(MITRE_techniques_overall):
            self._MITRE_techniques = pd.read_csv(MITRE_techniques_overall, sep=";")
        MITRE_techniques_per_tactic = "./data/techniques_frequency_per_tactic.json"
        if file_exists(MITRE_techniques_per_tactic):
            self._MITRE_techniques_per_tactic = pd.read_json(MITRE_techniques_per_tactic)

    def __MITRE_tactic_idx(self, tactic: str) -> int:
        """
        Return internal index of given MITRE :param:`tactic`
        :param str tactic: MITRE name of the tactic (MITRE ID, name, or codename)
        :return: internal index of the tactic
        """
        try:
            if tactic[0] == "T":
                return self._MITRE_tactic_ID.index(tactic)
            elif tactic[0].isupper():
                return self._MITRE_tactic_names.index(tactic)
            elif len(tactic) == 4:
                return self._internal_tactic_ID.index(tactic)
            else:
                return self._MITRE_tactic_codenames.index(tactic)
        except ValueError as e:
            raise KeyError("No MITRE tactic '" + tactic + "' found")

    def __parent_node(self, node: str) -> str:
        """ Return (name of) *a* parent of the given node """
        edges = list(self._AT.out_edges()(node))
        assert edges  # node_idx must have parents
        return edges[0][1]

    def __root_from_node(self, node: str) -> str:
        """ Find the AT root starting from the given node, with naive loop detection """
        AT = self._AT
        steps = 0
        while steps < self._num_nodes and list(AT.out_edges()(node)):  # this node has parents
            node = self.__parent_node(node)
            steps += 1
        if steps >= self._num_nodes:
            raise (Exception("Cyclic logic detected, aborting root-find cycle"))
        return node

    def __define_root(self) -> None:
        """ Define the AT root, i.e. the node with no parents """
        nodes = list(self._AT.nodes())
        assert (len(nodes) > 0)
        self._root = self.__root_from_node(nodes[0])

    ##########################################################################
    #  Setters  ##############################################################

    def add_BAS(self,
                ID: str,
                prob: float = 0,
                color: str = "#dddddd",
                **kwargs) -> None:
        """
        Add a BAS to the AT

        :param string ID: identifier, must be unique in the whole AT
        :param float prob: probability of execution
        :param string color: node color for AT plotting via :func:`show()`
        :param kwargs: (optional) comma-separated series of name=value pairs
        """
        if "alias" in kwargs.keys() and kwargs["alias"] in self._BASes.keys():
            raise AssertionError("Cannot add BAS: name '" + kwargs["alias"] + "' already exists")
        elif "alias" not in kwargs.keys() and ID in self._BASes.keys():
            raise AssertionError("Cannot add BAS: name '" + ID + "' already exists")
        assert (0 <= prob)
        assert (prob <= 1)
        if ID not in self._BASes.keys():
            # new node (else it's an alias to a pre-existent node)
            self._BASes[ID] = prob  # our reference
            self._AT.add_node(ID, type="BAS", prob=prob, shape="cylinder", **kwargs)  # internal ADT reference
            self._num_nodes += 1
            self._node_shapes += ["o"]
            self._node_colors += [color]

    @deprecated
    def add_BAS_MITRE_global(self,
                             name: str,
                             tech: str):
        """
        Add a BAS to the AT, mapped to a MITRE technique

        :param string name: identifier, must be unique in the whole AT
        :param string tech: MITRE technique associated to this attack (probability automatically computed)
        """
        if self._MITRE_techniques is None:
            raise (Exception("Cannot use method: no data available for MITRE techniques"))
        if name in self._BASes.keys():
            raise (AssertionError("Cannot add BAS: name '" + name + "' already exists"))
        df = self._MITRE_techniques[self._MITRE_techniques['target ID'] == tech]
        if len(df) == 0:
            raise KeyError("Cannot add BAS: MITRE technique '" + tech + "' not found")
        elif len(df) > 1:
            raise KeyError("Cannot add BAS: MITRE technique '" + tech + "' has multiple matches!??")
        self.add_BAS(name, float(df["prob"].iloc[0]), tech)

    def add_BAS_MITRE(self,
                      name: str,
                      tactic: str,
                      technique: str,
                      color: str = "#dddddd",
                      is_dummy: bool = False,
                      **kwargs) -> None:
        """
        Add a BAS to the AT, mapped to a MITRE technique whose probability
        (frequentist, per tactic) is automatically computed

        :param string name: alias of the node, must be unique in the whole AT
        :param string tactic: MITRE tactic associated to this attack (probability automatically computed)
        :param string technique: MITRE technique associated to this attack (probability automatically computed)
        :param string color: node color for AT plotting via :func:`show()`
        :param kwargs: (optional) comma-separated series of name=value pairs
        """
        if self._MITRE_techniques_per_tactic is None:
            raise (Exception("Internal error: no data available for MITRE techniques"))
        if name in self._BASes.keys():
            raise (AssertionError("Cannot add BAS: name '" + name + "' already exists"))
        # translate tactic name to indexed names
        tactic_ID = self._internal_tactic_ID[self.__MITRE_tactic_idx(tactic)]
        tactic_MITRE = self._MITRE_tactic_codenames[self.__MITRE_tactic_idx(tactic)]
        # look for probability in database
        try:
            prob = self._MITRE_techniques_per_tactic[tactic_MITRE][technique]
        except KeyError:
            prob = 0  # allow tactic×technique that doesn't exist in MITRE campaign
        finally:
            if math.isnan(prob) or is_dummy:
                prob = 0
                # # alternatively raise hell
                # raise (KeyError("Cannot add BAS: MITRE technique '{}' not found for tactic '{}'".
                #                 format(technique, tactic)))
        # if default colour but BAS not dummy, then "turn on this BAS" with colour red
        if color == "#dddddd" and not is_dummy:
            color = "#b70020"  # dark red          # "#ff0033"  # red
        # ID of BAS is the MITRE technique+tactic
        techtac = technique + "\n" + tactic_ID
        self._techtac[name] = techtac  # register alias
        if techtac in self._BASes.keys():
            print("Warning: BAS for '{}' already exists—'{}' will be aliased to that BAS".
                  format(techtac, name))
        self.add_BAS(ID=techtac, prob=prob, color=color, alias=name, tactic=tactic, technique=technique, **kwargs)

    def update_BAS(self,
                   name: str,
                   prob: float = 0,
                   **kwargs) -> None:
        """ Update the attributes of an existent BAS — See :fun:`add_BAS()` """
        if name not in self._BASes.keys():
            raise (AssertionError("Cannot update BAS: name '" + name + "' does not exist"))
        assert (0 <= prob)
        assert (prob <= 1)
        self._BASes[name] = prob  # our reference
        kwargs.update(prob=prob)
        self._AT.add_node(name, **kwargs)  # internal ADT reference

    def add_gate(self,
                 name: str,
                 gtype: str,
                 color: str = "#dddddd",
                 **kwargs) -> None:
        """
        Add a gate to the AT

        :param string name: identifier, must be unique in the whole AT
        :param string gtype: gate type, must be one of ["OR", "AND"]
        :param string color: node color for AT plotting via :func:`show()`
        :param kwargs: (optional) comma-separated series of name=value pairs
        """
        if name in self._gates.keys():
            raise (AssertionError("Cannot add gate: name '" + name + "' already exists"))
        if gtype not in self._gates_types:
            raise (AssertionError("Cannot add gate: type '" + gtype + "' is not one of " + str(self._gates_types)))
        self._gates[name] = gtype  # our reference
        self._AT.add_node(name,
                          type=gtype,
                          shape=self._gates_shapes2[gtype],
                          style="filled",
                          **kwargs)  # internal ADT reference
        self._num_nodes += 1
        self._node_shapes += [self._gates_shapes[gtype]]
        self._node_colors += [color]

    def update_gate(self,
                    name: str,
                    gtype: str,
                    **kwargs) -> None:
        """ Update the attributes of an existent gate — See :fun:`add_gate()` """
        if name not in self._gates.keys():
            raise (AssertionError("Cannot update gate: name '" + name + "' does not exist"))
        if gtype not in self._gates_types:
            raise (AssertionError("Cannot add gate: type '" + gtype + "' is not one of " + str(self._gates_types)))
        self._gates[name] = gtype  # our reference
        kwargs.update(type=gtype)
        self._AT.add_node(name, **kwargs)  # internal ADT reference

    def add_logic(self,
                  child: str,
                  parent: str):
        """
        Set a logical connection between nodes in the AT

        :param string child: name of the child BAS/gate
        :param string parent: name of the parent gate
        """
        assert (parent in self._gates.keys())
        if child in self._techtac.keys():
            child = self._techtac[child]  # child refers to this techtac node
        if child not in self._BASes.keys() and child not in self._gates.keys():
            raise (AssertionError("Node name " + child + " not found"))
        self._AT.add_edge(child, parent)

    def connect(self,
                from_id: str,
                to_id: str):
        """ See :func:`add_logic()` """
        self.add_logic(from_id, to_id)

    def add_gate_logic(self,
                       name: str,
                       gtype: str,
                       children: [str],
                       color: str = "#dddddd",
                       **kwargs):
        """ Combination of :func:`add_gate()` and :func:`add_logic()` """
        self.add_gate(name, gtype, color, **kwargs)
        for child in children:
            self.add_logic(child, name)

    ##########################################################################
    #  Getters  ##############################################################

    def BAS(self, name: str):
        """ Get data of BAS with this name == (prob, descr) """
        return self._BASes[name]

    def get_BAS(self, name: str):
        """ See :func:`BAS()` """
        return self.BAS(name)

    def gate(self, name: str):
        """ Get data of gate with this name == (type, descr) """
        return self._gates[name]

    def get_gate(self, name: str):
        """ See :func:`gate()` """
        return self.gate(name)

    def BASes(self):
        """ Get all BASes — See :func:`BAS()` """
        return self._BASes

    def get_BASes(self):
        """ See :func:`BASes()` """
        return self.BASes()

    def gates(self):
        """ Get all gates — See :func:`gate()` """
        return self._gates

    def get_gates(self):
        """ See :func:`gates()` """
        return self.gates()

    def get_root(self):
        """ Root node of this AT, viz. the gate with no parents or a BAS """
        self.__define_root()
        return self._root

    def __nodes(self):
        """ Get (internal references to) all BASes and gates """
        return self._AT.nodes()

    def __edges(self):
        """ Get (internal references to) all connections between BASes and gates """
        return self._AT.edges()

    ##########################################################################
    #  Utils  ################################################################

    def is_static(self, verbose: bool = False) -> bool:
        """ Is this a static Attack Tree? """
        for name, gtype in self.gates().items():
            if gtype not in ["OR", "AND"]:
                if verbose:
                    print("WARNING: AT is not static! Gate \"{}\" has type {}".
                          format(name, gtype))
                return False
        return True

    def is_well_defined(self, verbose: bool = False) -> bool:
        """
        Is the internal structure a single-rooted connected DAG,
        whose every intermediate node has at least two children?
        NOTE: IOU an efficient implementation
        """
        isAttackTree = True
        self.__define_root()
        for n in self.__nodes():
            if self.__root_from_node(n) != self._root:
                isAttackTree = False
                if verbose:
                    print("WARNING: AT is not connected! Root '{}' isn't reachable from node '{}'".
                          format(self._root, n))
                break
            parents = self._AT.out_edges(n)
            children = self._AT.in_edges(n)
            if len(parents) == 0 and n != self._root:
                isAttackTree = False
                if verbose:
                    print("WARNING: AT is not a single-rooted DAG! Two roots found: '{}' and '{}'".
                          format(self._root, n))
                break
            elif len(children) == 1:
                isAttackTree = False
                if verbose:
                    print("WARNING: AT is not an Attack Tree! Node '{}' has a single child ('{}')".
                          format(n, children[0][1]))
                break
        return isAttackTree

    # def show(self, node_colors: [str] = [], fig_size: [int, int] = (7, 7)) -> None:
    def show(self,
             size: [int, int] = (7, 7),
             layout: str = "twopi",
             fname: str = "") -> None:
        """
        Display graphical representation of the *graph* (no AT shapes)

        :param string size: width,height of image to produce
        :param string layout: Graphviz layout to use, e.g. "twopi" for radial graphs, "dot" otherwise
        :param string fname: if non-empty, don't show but rather save file "fname" with image (extension included)
        """
        AT = self._AT
        # pos = nx.nx_pydot.pydot_layout(AT)
        pos = nx.nx_agraph.graphviz_layout(AT, prog=layout)
        plt.figure(figsize=size)
        # Nodes
        nx.draw_networkx_nodes(AT, pos, node_color=self._node_colors)  #, node_shape=self._node_shapes)
        # Edges
        nx.draw_networkx_edges(AT, pos)  # , connectionstyle="arc3,rad=0.06")
        # Labels
        nx.draw_networkx_labels(AT, pos, font_size=6)
        # Plot
        plt.axis("equal")
        if len(fname) > 0:
            print(f"Saving AT image in {fname}")
            plt.savefig(fname)
        else:
            plt.show()

    def evil_twin(self, verbose: bool = False):
        """ See :func:`create_static_copy()` """
        return self.create_static_copy(verbose)

    def create_static_copy(self, verbose: bool = False):
        """ Static-AT (deep) copy of this AT, replacing SANDs for ANDs """
        twin = copy.deepcopy(self)
        for g in twin._gates:
            if twin.gate(g) == "SAND":
                twin.update_gate(g, "AND")
                if verbose:
                    print("INFO: SAND gate \"{}\" turned into AND".format(g))
        assert (twin.is_static(verbose))
        return twin

    def compute_metric(self,
                       metric: str,
                       verbose: bool = False,
                       force: bool = False) -> float:
        """
        Compute requested metric on AT, which must be one of :mem:`METRICS`.
        See Lopuhaä-Zwakenberg et al. (2023), DOI: 10.1109/TDSC.2022.3215752
        :param str metric: Metric to compute, _BASes must have homonymous attribute values_
        :param bool verbose: https://t.ly/YAEvo
        :param bool force: Attempt computation without checking AT consistency
        :return: Metric value of the top node of the AT
        """
        if metric not in METRICS:
            raise (AssertionError("Unsupported metric \"{}\": must be one of {}".
                                  format(metric, METRICS)))
        self.__define_root()
        if not force:
            assert (self.is_well_defined(verbose))
        assert (self.is_static(verbose))
        return AT_metric(self._AT, self._root, metric, verbose)
