#!/usr/bin/python3 -u
#
#  AttackTree_algorithms.py
#
#  Efficient and generic algorithms that compute metrics from Attack Trees
#  See Lopuhaä-Zwakenberg et al. (2023), DOI: 10.1109/TDSC.2022.3215752
#
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
#
#  Distribution license: GPLv3
#                        https://www.gnu.org/licenses/gpl-3.0.en.html
#  Copyright 2024–
#  Authors:
#  - Vasko Pirinski <v.pirinski@student.utwente.nl> (University of Twente, NL)
#  - Carlos E. Budde <carlosesteban.budde@unitn.it> (University of Trento, IT)
#  - Stefano M. Nicoletti <s.m.nicoletti@utwente.nl> (University of Twente, NL)
#
# ------------------------------------------------------------------------------

import networkx as nx
from math import inf
from functools import reduce
from dd import autoref as _bdd  # https://github.com/tulip-control/dd

#   ^^^ May need a Python venv, e.g. for dependencies such as astutils
#       Run in Jupyer notebook via pipenv: [stackoverflow.com/a/55081892]


DISJUNCTION = r"\/"  # ∨ (aka logical OR)
CONJUNCTION = "/\\"  # ∧ (aka logical AND)
BEXPR_OPEN = "( "
BEXPR_CLOSE = " )"


def __literal(BAS_name: str) -> str:
    """
    Transform BAS name into BDD-compliant literal form
    :param BAS_name: Name of the BAS
    :return: Name of the literal corresponding to that BAS
    """
    return ("_".join(BAS_name.split())
            .replace("&", "and")
            .replace('|', "or")
            .replace('+', "plus")
            .replace('*', "times")
            .replace('-', "_")
            .replace('/', "_")
            .replace('\\', "_")
            .replace('.', "_")
            .replace(',', "_")
            .replace('$', "_")
            )

def cATM_MITRE_template_prob_AND(x: float, y: float) -> float:
    """
    Propagate probabilities of BASes for an AND gate
    assuming the structure of the MITRE attack tree template,
    where ANDs could have BASes "turned-off", and then we
    want to continue computation without killing with a 0.
    """
    # x and y should be probability values
    assert (0 <= x <= 1)
    assert (0 <= y <= 1)
    if x + y == 0:
        # if both are zero then propagate neutral of AND
        return 1.0
    elif x == 0 or y == 0:
        # if a child is zero don't kill product, but continue with rest
        return max(x, y)
    else:
        # true behaviour of AND gate
        return x*y

def cATM_MITRE_template_prob_OR(x: float, y: float) -> float:
    """
    Propagate probabilities of BASes for an OR gate
    assuming the structure of the MITRE attack tree template,
    where ORs could have all BASes "turned-off", and then we
    want to continue computation without killing the AND with a 0.

    Note: assumes no BAS (i.e. MITRE technique) has probability == 1
          ^^^ this is reflected in the "return min(x,y)" below
    """
    # x and y should be probability values
    assert (0 <= x <= 1)
    assert (0 <= y <= 1)
    if x + y == 0:
        # if both are zero then propagate neutral of AND
        return 1.0
    elif 1 == x or 1 == y:
        # 1 caused by the case above: return the lowest (see Note in docstring)
        return min(x, y)
    else:
        # true behaviour of OR gate
        return max(x, y)

__METRICS = {
    #  name: (operator AND,
    #         operator OR,
    #         neutral AND,
    #         neutral OR)
    "cost": (lambda x, y: x + y,
             min,
             0,
             inf),
    "prob": (lambda x, y: x * y,  # max probability LOAD
             max,                 # max probability LOAD
             1.0,
             0.0),
    "cATM": (cATM_MITRE_template_prob_AND,  # equivalent to max probability
             cATM_MITRE_template_prob_OR,   # equivalent to max probability
             1.0,
             0.0),
}

# Metrics offered for computation
METRICS = list(__METRICS.keys())


def AT_metric(AT: nx.DiGraph,
              AT_root,
              metric: str,
              verbose: bool = False):
    """
    Compute specified AT metric of top-level attack

    - *AT structure:* Works for both tree- and DAG-structured ATs
    - *Gates types:*  It only supports static gates (AND and OR), not dynamic gates such as SANDs

    :param networkx.DiGraph AT: Attack Tree, as a single-rooted directed acyclic graph
    :param str AT_root: root of AT (supports networkx-DiGraph nodes only)
    :param str metric: Name of the attribute that defines the metric to quantify
    :param bool verbose: Print extra data, e.g. the Boolean formula of the AT

    See Lopuhaä-Zwakenberg et al. (2023), DOI: 10.1109/TDSC.2022.3215752
    """
    assert (AT is not None)
    assert (AT_root is not None)
    assert (metric in METRICS)

    attribute = "prob" if metric == "cATM" else metric  # cATM metric uses "prob" attribute of BASes

    # Distinguish case: if DAG use BDD transformation, else do bottom-up directly on AT
    if len(nx.cycle_basis(AT.to_undirected())) <= 0:
        # No undirected cycles found => AT is a proper tree => do that bottom
        if verbose:
            print(f"AT is a proper tree: using bottom-up algorithm")
        metric_value = AT_bottom_up(AT, AT_root, metric, attribute, verbose)
    else:
        # Undirected cycles found => AT is a "proper" DAG => use BDD transformation
        if verbose:
            print(f"AT is a DAG: using BDD transformation")
        bdd, bdd_root, values = AT_to_quantified_BDD(AT, AT_root, attribute, verbose)
        metric_value = BDD_bottom_up(bdd, bdd_root, values, metric, attribute)

    if verbose:
        print(f"Top-level attack \"{metric}\" value: {metric_value}")
    return metric_value

def AT_bottom_up(AT: nx.DiGraph,
                 AT_root,
                 metric: str,
                 attribute: str,
                 verbose: str = False):
    """
    Compute a metric by traversing the AT bottom-up.

    NOTE: algorithm assumes that the direction of the edges goes from children to parents,
          e.g. from BASes to gates
    NOTE: complexity is linear in the number of nodes of AT (starting from AT_root)

    :param networkx.DiGraph AT: Attack Tree, as a single-rooted directed acyclic graph
    :param str AT_root: root of AT (supports networkx-DiGraph nodes only)
    :param str metric: Name of the attribute that defines the metric to quantify
    :param str attribute: Attribute of the BASes that correspond to metric, if empty defaults to metric
    :param bool verbose: Print the BASes attribute values

    :return: Metric value of the AT, starting from given root node
    """
    children = AT.pred[AT_root]
    if len(children) <= 0:
        # leaf == BAS -> return its attribution
        attr = AT.nodes[AT_root][attribute]
        if verbose and attr > 0:
            node = AT_root.replace("\n", "_").replace(".","_")
            print(f"\t{node}: {attr}")
        return attr
    else:
        # gate -> aggregate the metric of its children using the right operator
        agg = __METRICS[metric][0] if AT.nodes[AT_root]["type"] == "AND" else __METRICS[metric][1]
        return reduce(agg, [AT_bottom_up(AT, child, metric, attribute, verbose) for child in children])

def BDD_bottom_up(bdd: _bdd,
                  root,
                  values: dict,
                  metric: str,
                  attribute: str = ""):
    """
    Compute a metric by traversing the BDD with the given aggregation function
    and true/false values as base cases
    :param _bdd bdd: BDD representing an Attack Tree
    :param root: Root node of the BDD in the current recursion step
    :param dict values: Attribute values of the BDD literals (i.e. the AT BASes)
    :param str metric: Metric to compute, see :mem:`__METRIC`
    :param str attribute: Attribute of the BASes that correspond to metric, if empty defaults to metric
    :return: Metric value of the BDD starting from given root node
    """
    attribute = attribute if len(attribute) > 0 else metric
    if root.node == -1:
        return __METRICS[metric][3]  # neutral OR == base case FALSE
    elif root.node == 1:
        return __METRICS[metric][2]  # neutral AND == base case TRUE
    else:
        return __METRICS[metric][1](  # operator OR
            BDD_bottom_up(bdd, root.low, values, metric),
            __METRICS[metric][0](  # operator AND
                BDD_bottom_up(bdd, root.high, values, metric),
                values[root.var])
        )


def AT_to_quantified_BDD(AT: nx.DiGraph,
                         root,
                         attribute: str,
                         verbose: bool = False):
    """
    Generate a BDD that represents the given Attack Tree,
    where the BASes are interpreted as literals decorated with attribute values
    :param networkx.DiGraph AT: Attack Tree, represented by a :mem:`networkx.DiGraph`
    :param int root: Root of AT, indexable as node of AT
    :param str attribute: Name of the attribute that defines the metric to quantify
    :param bool verbose: Print the Boolean formula for the AT, and the BASes attribute values
    """
    bexpr, values = AT_to_QBF(AT, root, attribute)
    if verbose:
        print("AT Boolean formula:\n\t" + bexpr)
        print("Attributes of the BASes:")
        for BAS, attr in values.items():
            print("\t{}: {}".format(BAS, str(attr)))
    bdd = _bdd.BDD()
    bdd.declare(*[__literal(x) for x in AT.nodes()])
    u = bdd.add_expr(r"{}".format(bexpr))
    bdd.collect_garbage()
    bdd.reorder()   # FIXME: is this needed to get a smaller BDD???
    return bdd, u, values


def AT_to_QBF(AT: nx.DiGraph,
              root,
              attribute: str,
              values: dict[str:float] = {}) -> (str, dict[str:float]):
    """
    Attributed Attack Tree to Quantified Boolean Formula

    Compose the Boolean formula represented by a static Attack Tree,
    interpreting the BASes as literals that are quantified by the given attribute values

    :param networkx.DiGraph AT: Attack Tree, represented by a :mem:`networkx.DiGraph`
    :param int root: Root of AT, indexable as node of AT
    :param str attribute: Name of the attribute that defines the metric to quantify, e.g. "cost"
    :param dict[str,float] values: Container for the attribute values of the literals

    :return: (x,_): Boolean formula as string;
             (_,y): metric quantities of the literals
    """
    assert (root is not None)
    root_data = AT.nodes()[root]
    children = [child for child, _ in AT.in_edges(root)]
    assert (len(children) == 0 or len(children) > 1)  # BAS or gate with child*ren*

    if len(children) == 0:
        lit = __literal(root)
        return lit, values | {lit: root_data[attribute]}
    else:
        # Gate (i.e. recursion == logical operation)
        bexpr = BEXPR_OPEN
        gate = root_data['type']
        if gate == "OR":
            OPERATOR = DISJUNCTION
        elif gate == "AND":
            OPERATOR = CONJUNCTION
        else:
            raise (TypeError("Unsupported gate type: {} (gate {} \"{}\")".
                             format(gate, root, AT.nodes()("root"))))
        for child in children:
            if bexpr != BEXPR_OPEN:
                bexpr += " " + OPERATOR + " "
            expr, values = AT_to_QBF(AT, child, attribute, values)
            bexpr += expr
        bexpr += BEXPR_CLOSE
        return bexpr, values
