#!/usr/bin/env python3 -u
#
#  MITRE_AT_templetisator.py
#
#  Generator of MITRE attack tree templates
#  See Nicoletti et al. (to appear)
#
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
#
#  Distribution license: GPLv3
#                        https://www.gnu.org/licenses/gpl-3.0.en.html
#  Copyright 2024–
#  Authors:
#  - Carlos E. Budde <carlosesteban.budde@unitn.it> (Università di Trento, IT)
#  - Stefano Nicoletti <s.m.nicoletti@utwente.nl) (Universiteit Twente, NL)
#
# ------------------------------------------------------------------------------

import json
import sys
from os import scandir
import pandas as pd

COLOR_AND_GATE = "#cc79a7"  # "#cc6677"
COLOR_OR_GATE = "#88ccee"
MITRE_subtechniques = {}  # populated below


def sane_name(name: str):
    # sanitise name for code reflection
    return name.replace(" ", "") \
        .replace("/", "") \
        .replace("\\", "") \
        .replace("-", "") \
        .replace(".", "")


def map_MITRE_techniques(df_techniques: pd.DataFrame) -> dict:
    """
    Create a mapping from "high-level" techniques to their subtechniques
    :param pd.DataFrame df_techniques: Pandas dataframe that contains at least the 'ID' column of the MITRE
        enterprise-attack techniques Excel spreadsheet (see https://attack.mitre.org/resources/attack-data-and-tools/)
    :return: Dictionary whose keys are "high-level" techniques, and whose corresponding values are their subtechniques
    """
    techniques = list(df_techniques['ID'].drop_duplicates().sort_values())
    N = len(techniques)
    MITRE_techniques_map = {}
    for i, t in enumerate(techniques):
        if len(t) == 5:
            # 't' is a new ("high-level") technique
            MITRE_techniques_map[t] = set()
            t_next = techniques[(i + 1) % N]
            if t != t_next[:5]:
                MITRE_techniques_map[t].add(t)  # no subtechniques: consider this technique as its own subtechnique
        else:
            # 't' should be a subtechnique of the previous technique, because the techniques list is sorted
            technique = t[:5]
            assert (technique == techniques[(i - 1) % N][:5])
            MITRE_techniques_map[technique].add(t)
    assert len(MITRE_techniques_map) > 0, "No techniques found"
    return MITRE_techniques_map


def normalise_techniques(df: pd.DataFrame,
                         column: str = 'techniqueID') -> pd.DataFrame:
    """
    Normalise `column` data to be able to identify sub-techniques per tactic, considering three cases:

    - **Fine-grained data**: a tactic has only certain subtechniques of a given high-level technique
       * then interpret each such subtechnique as itself
    - **Coarse-grained data**: a tactic has a certain high-level technique but none of its subtechniques
       * then interpret that each subtechnique took case (worst-case scenario)
    - **Mixed data**: a tactic has a high-level technique, and some of its subtechniques as well
       * use fine-grained data (most informative), disregarding the high-level technique

    **NOTE:** `df` should be fine-grained enough to be able to tell
    the fine, coarse, and mixed cases. If repetitions in the `column` are
    a mistake, then this condition is most likely satisfied.

    :param pd.DataFrame df: DataFrame to normalise
    :param string column: Column of 'df' with MITRE (techniques and) subtechniques
    :return: Copy of `df` normalised as per data in `column`
    """
    assert len(df[column]) == len(df[column].unique()), "Data isn't fine-grained enough"
    dfn = df.copy()
    techniques = dfn[column]
    for t in [t for t in techniques if len(t) == 5]:
        # technique 't' is high-level, i.e. not a subtechnique name
        if t not in MITRE_subtechniques.keys():
            continue  # wat
        subtechniques = MITRE_subtechniques[t]
        if t in subtechniques:
            # 't' is its own subtechnique == fine case
            continue
        if [st for st in techniques if len(st) > 5 and st.startswith(t)]:
            # subtechniques of 't' are present == mixed case
            dfn.drop(dfn[dfn[column] == t].index, inplace=True)  # drop the technique
        else:
            # 't' stands alone == coarse case
            dfn[column] = dfn[column].apply(lambda x: list(subtechniques) if x == t else x)  # add all subtechniques
    assert len(dfn) > 0
    return dfn.explode(column, ignore_index=True).copy()  # kaboom!


def add_relevant_techniques(tactic: str, BASes: str) -> str:
    return "for i in range(len(" + BASes + ")):" + "\n" \
        + "\tif " + BASes + f"[i][-1] in campaign_tactech[\"{tactic}\"]:\n" \
        + "\t\t# This is a technique from our campaign (or we're generating the full MITRE AT template)\n" \
        + "\t\tAT.add_BAS_MITRE(*" + BASes + "[i])\n" \
        + "\telse:\n" \
        + "\t\t# This is a technique outside our campaign\n" \
        + "\t\tAT.add_BAS_MITRE(*" + BASes + "[i], is_dummy=True)\n"


def MITRE_AT_template(campaign: str = "",
                      difficulty: str = "default",
                      MITRE_CAMPAIGNS_DIR: str = "./data/MITRE_campaigns/",
                      MITRE_MATRIX: (str, str) = ("./data/enterprise-attack-v14.1-matrices-ordered.ods", "formatted_matrix"),
                      MITRE_TECHNIQUES: (str, str) = ("./data/enterprise-attack-v14.1-techniques.xlsx", "techniques")) -> str:
    """
    Create a Python script that, when executed, creates a MITRE attack tree template for the requested campaign,
    by generating the complete MITRE AT template and turning off (via cATM) the BASes not used in the campaign.
    If no campaign given then generate the complete MITRE AT template, i.e. all BASes are turned on.

    :param str campaign: MITRE campaign codename, e.g. "C0014" for Wocao
    :param str difficulty: how strict is the interpretation of the campaign data from MITRE, can be:
                           "hard", i.e. campaign succeeds iff exact (sub)techniques from MITRE are used;
                           "easy", i.e. campaign succeeds iff at least one subtechnique per tactic is used;
                           "default", i.e. at least one subtechnique per technique, all techniques per tactic.
    :param str MITRE_CAMPAIGNS_DIR: filesystem directory where to search for the JSON file that describes the
                                    techniques (per tactic) used in the requested campaign,
                                    as available in the ATT&CK navigators ("download single layer as json"),
                                    e.g. https://mitre-attack.github.io/attack-navigator//#layerURL=https%3A%2F%2Fattack.mitre.org%2Fcampaigns%2FC0014%2FC0014-enterprise-layer.json
    :param str MITRE_MATRIX: MITRE enterprise-attack matrices Excel spreadsheet,
                             as available at https://attack.mitre.org/resources/attack-data-and-tools/
    :param str MITRE_TECHNIQUES: MITRE enterprise-attack techniques Excel spreadsheet,
                                 as available at https://attack.mitre.org/resources/attack-data-and-tools/
    """
    global MITRE_subtechniques
    assert difficulty in ["hard", "easy", "default"], "RTFM"
    # Load MITRE techniques
    df = pd.read_excel(MITRE_TECHNIQUES[0], sheet_name=MITRE_TECHNIQUES[1])  # requires package python-openpyxl
    df_techniques = df[['ID', 'name']].drop_duplicates().sort_values(by=['ID', 'name'])
    MITRE_subtechniques = map_MITRE_techniques(df_techniques)
    # Load MITRE matrix
    df_matrix = pd.read_excel(MITRE_MATRIX[0], sheet_name=MITRE_MATRIX[1])  # requires package python-openpyxl
    df_matrix["childless_technique"] = (
            (df_matrix["only_technique"] != df_matrix.shift(1)["only_technique"]) &
            (df_matrix["only_technique"] != df_matrix.shift(-1)["only_technique"])
    )
    df_matrix["technique"] = df_matrix[["only_technique", "only_subtechnique"]].apply(
        lambda row: row.iloc[0] if pd.isna(row.iloc[1]) else ": ".join(row.values.astype(str)), axis=1)
    MITRE_TACTICS = set(df_matrix[df_matrix['tactic'].notna()]['tactic'])
    # Glorious join
    df_glorrrious = df_matrix.join(df_techniques.set_index("name"), on="technique")
    df_glorrrious = df_glorrrious[~pd.isna(df_glorrrious["tactic"])]
    # Load the campaign specified by the user, if any
    if len(campaign) > 0:
        campaign_json = ""
        for JSON in scandir(MITRE_CAMPAIGNS_DIR):
            if JSON.path.casefold().find(campaign.casefold()) >= 0:
                campaign_json = JSON.path
                break
        assert len(campaign_json) > 0, f"No JSON file found in {MITRE_CAMPAIGNS_DIR} for requested campaign {campaign}"
        # Organise techniques per tactic
        dfs = []
        with open(campaign_json) as f:
            data = json.load(f)
            df = pd.DataFrame.from_records(data['techniques'])
            assert ('tactic' in df.columns)
            assert ('techniqueID' in df.columns)
            # normalise techniques on each tactic
            for tactic in list(df['tactic'].unique()):
                dft = df[df['tactic'] == tactic]
                dfs.append(normalise_techniques(dft[['techniqueID', 'tactic']]))
        df_campaign = pd.concat(dfs)
    else:
        # generate the full MITRE AT template
        df_campaign = df_glorrrious[['tactic', 'ID']].rename(columns={'ID': 'techniqueID'})

    # Create dict from DataFrame, where keys are tactics and values are sets of techniques
    campaign_tactech = dict()
    CAMPAIGN_TACTICS = set(df_campaign['tactic'])
    for tactic in CAMPAIGN_TACTICS:
        campaign_tactech[tactic] = set(df_campaign[df_campaign['tactic'] == tactic]['techniqueID'])
    for tactic in MITRE_TACTICS - CAMPAIGN_TACTICS:
        campaign_tactech[tactic] = set()

    # Create program code
    program = "#!/usr/bin/env python3\n\n"
    program += "import pygraphviz\n"
    program += "from AttackTree import AttackTree" + "\n"
    program += "\n"
    program += "AT = AttackTree()" + "\n"
    program += "\n"
    program += "campaign_tactech = " + campaign_tactech.__str__()
    program += "\n"
    tactics = []
    # for tactic in campaign_tactech.keys():  # include only the campaign tactics in template
    for tactic in MITRE_TACTICS:              # include all 14 MITRE tactics in template
        program += "\n# - - - - - - - - - - - - - - - - - -" + "\n"
        program += "# Tactic: " + tactic + "\n" + "\n"
        df_tactic = df_glorrrious[df_glorrrious["tactic"] == tactic]
        df_childless = df_tactic[df_tactic["childless_technique"]]
        df_childfull = df_tactic[~df_tactic["childless_technique"]]
        subtrees = []
        # For techniques with children create an OR/AND gate, where subtechniques are leaves
        for tech in set(df_childfull["only_technique"]):
            program += "# " + tech + "\n"
            subtree = "_".join(list(map(sane_name, [tactic, tech])))
            # Gather subtechniques of this technique
            BASes = "BASes_" + subtree
            program += BASes + " = [" + "\n"
            df_tech = df_childfull[df_childfull["only_technique"] == tech]
            for i in range(len(df_tech)):
                program += "\t" + str(tuple(df_tech.iloc[i][["technique", "tactic", "ID"]])) + "," + "\n"
            program += "]" + "\n"

            # Add all those subtechniques (if present in campaign) as BASes to the AT
            program += add_relevant_techniques(tactic, BASes)

            # Add the OR/AND gate representing the technique
            if difficulty == "hard":
                # exact reproduction of data in MITRE is requested
                program += "\nAT.add_gate_logic(\"" + subtree + "\", \"AND\", color=\"" + COLOR_AND_GATE + "\", children=[ x[0] for x in " + BASes + "])" + "\n"
            elif difficulty == "easy":
                # at least one subtechnique per technique is requested
                program += "\nAT.add_gate_logic(\"" + subtree + "\", \"OR\", color=\"" + COLOR_OR_GATE + "\", children=[ x[0] for x in " + BASes + "])" + "\n"
            else:  # difficulty == "default"
                # by default we request at least one subtechnique per technique
                program += "\nAT.add_gate_logic(\"" + subtree + "\", \"OR\", color=\"" + COLOR_OR_GATE + "\", children=[ x[0] for x in " + BASes + "])" + "\n"
            program += "\n"
            subtrees.append("\"" + subtree + "\"")
        # For techniques without children, add them as direct leaves of the tactic
        program += "# direct BASes" + "\n"
        # Gather all techniques without a subtechnique
        BASes = "BASes_" + sane_name(tactic) + "_childless"
        program += BASes + " = [" + "\n"
        for i in range(len(df_childless)):
            program += "\t" + str(tuple(df_childless.iloc[i][["technique", "tactic", "ID"]])) + "," + "\n"
        program += "]" + "\n"

        # Add all those techniques as BASes to the AT
        program += add_relevant_techniques(tactic, BASes)

        # Add the OR/AND gate representing the tactic
        if difficulty == "hard":
            # exact reproduction of data in MITRE is requested
            program += "\nAT.add_gate_logic(\"" + tactic + "\", \"AND\", color=\"" + COLOR_AND_GATE + "\", children=[" + ",".join(
                subtrees) + "]+[x[0] for x in " + BASes + "])" + "\n"
        elif difficulty == "easy":
            # at least one technique per tactic is requested
            program += "\nAT.add_gate_logic(\"" + tactic + "\", \"OR\", color=\"" + COLOR_OR_GATE + "\", children=[" + ",".join(
                subtrees) + "]+[x[0] for x in " + BASes + "])" + "\n"
        else:  # difficulty == "default"
            # by default, we request all techniques listed for this tactic
            program += "\nAT.add_gate_logic(\"" + tactic + "\", \"AND\", color=\"" + COLOR_AND_GATE + "\", children=[" + ",".join(
                subtrees) + "]+[x[0] for x in " + BASes + "])" + "\n"

        program += "\n"
        tactics.append("\"" + tactic + "\"")

    # Connect all tactics at the TLE
    program += "# TLE connecting all tactics" + "\n"
    program += "\nAT.add_gate_logic(\"glorrrious_TLE\", \"SAND\", color=\"" + COLOR_AND_GATE + "\", children = [" + ",".join(
        tactics) + "])" + "\n"

    # Embed some executions of interest, e.g. computation of metrics and display of AT
    program += "\n"
    program += "assert (AT.is_well_defined(verbose=True))\n"
    program += "full_AT = AT.evil_twin(verbose=True)\n"
    program += "assert (full_AT.is_static(verbose=True))\n"
    program += "assert (full_AT.is_well_defined(verbose=True))\n"
    program += "\n"
    program += "print(\"Computing AT probability via cATM...\")\n"
    program += "print(\"AT cATM: \" + str(full_AT.compute_metric(\"cATM\", verbose=True)))\n"
    program += f"AT.show(size=(111,111), fname=\"MATT_{difficulty}_{campaign}.pdf\")\n"

    return program


if __name__ == "__main__":
    MITRE_AT_template(*sys.argv[1:4])
